/*

HashCart-XT secure hash function based on XTEA-3 (cipher extensions by Tom St Denis)

(c) 2009 Idea of hash design and implementation by Alexander Myasnikov

e-mail: alexanderwdark@ya.ru

 */

#include <stdlib.h>
#include <stdio.h>
#include <mem.h>


#define HARDW_ROTATE

typedef signed long int s32;
typedef unsigned long int u32;
typedef unsigned char u8;

/*

Stores current hash value

 */

union key_union
  {
    u32 b32[8];
    u8 b8[32];
  };

union key_union key[2];

const u32 ITERATIONS = 32;


#ifdef HARDW_ROTATE
#define rol(N, R) _lrotl(N, R)
#else

u32 rol(u32 base, u32 shift)
{
  u32 res;
  shift &= 0x1F;
  res = (base << shift) | (base >> (32 - shift));
  return res;
}
#endif


/*

XTEA-3 encryption routine without KW

 */

void Transform(u32 keynum, u32 * plain, u32 * out)
{
  u32 a, b, c, d, sum, r, t;

  sum = 0;

  a = plain[0];
  b = plain[1];
  c = plain[2];
  d = plain[3];

  for (r = 0; r != ITERATIONS; r++)
    {
      a = a + (((b << 4) + rol(key[keynum].b32[(sum % 4) + 4], b)) ^
               (d + sum) ^ ((b >> 5) + rol(key[keynum].b32[sum % 4], b >> 27)));
      sum = sum + 0x9E3779B9;
      c = c + (((d << 4) + rol(key[keynum].b32[((sum >> 11) % 4) + 4], d)) ^
               (b + sum) ^ ((d >> 5) + rol(key[keynum].b32[(sum >> 11) % 4],
                                           d >> 27)));

      t = a;
      a = b;
      b = c;
      c = d;
      d = t;
    }

  out[0] = a;
  out[1] = b;
  out[2] = c;
  out[3] = d;
}

/*

Blocks counter for each update step

 */

u32 blocks = 0, currentblock = 0, blocksleft = 0;

/*

Checksums of all data and last value for key setup, used in update step

 */

u32 crc_tab[256], crc32 = 0xFFFFFFFF, lastXOR = 0xeb86d391;

/*

Simple CRC32 table initialization, may be replaced by 256 constant values

 */

void crc32gentab(void)
{
  u32 crc, poly;
  s32 i, j;

  poly = 0xEDB88320L;

  for (i = 0; i != 256; i++)
    {
      crc = i;
      for (j = 8; j != 0; j--)
        {
          if (crc & 1)
            {
              crc = (crc >> 1) ^ poly;
            }
          else
            {
              crc >>= 1;
            }
        }
      crc_tab[i] = crc;
    }
}

/*

Hash initializaion, use it once for each file or message

if you need verification, submit 256-bit key and set usekey to 1

 */

void InitHash(u32 datalen, u32 usekey, u32 * userkey)
{

  u32 i;

  if (datalen < 16)
    {
      blocks = 1;
    }
  else
    {
      blocks = (datalen + (datalen % 16)) / 16;
    }

  currentblock = 0;

  blocksleft = blocks;

  /*

  Magic words for first key

   */

  key[0].b32[0] = 0xd76aa478;
  key[0].b32[1] = 0xe8c7b756;
  key[0].b32[2] = 0x242070db;
  key[0].b32[3] = 0xc1bdceee;

  if (usekey == 1)
    {

      for (i = 0; i != 8; i++)
        key[1].b32[i] = userkey[i];

    }

  else
    {

      /*

      Magic words for second key and data length as last 32-bit value

       */

      key[1].b32[0] = 0xf57c0faf;
      key[1].b32[1] = 0x4787c62a;
      key[1].b32[2] = 0xa8304613;
      key[1].b32[3] = 0xfd469501;
      key[1].b32[4] = 0x698098d8;
      key[1].b32[5] = 0x8b44f7af;
      key[1].b32[6] = 0xffff5bb1;
      key[1].b32[7] = datalen;
    }

  crc32gentab();

  crc32 = 0xFFFFFFFF;
  lastXOR = 0xeb86d391;

}

/*

Hash calculation routine, submit 16 bytes (128 bit) value

if data is too short, pad it with zeroes

byte 15 = initial length
byte 14 = number of pad zeroes
0..13 = if empty, must be zero

 */

void Update(u8 * buffer)
{

  u32 i, *tmp;

  currentblock++;
  blocksleft--;

  key[0].b32[4] = currentblock;
  key[0].b32[5] = blocksleft;
  key[0].b32[6] = crc32 & 0xffffffff;
  key[0].b32[7] = lastXOR;

  /*

  Transform non-data based key value by compression function using default (or user-submited) key

   */

  Transform(1, (u32*) & key[0].b32[4], (u32*) & key[0].b32[4]);

  lastXOR = 0;

  tmp = (u32*) & buffer[0];

  for (i = 0; i != 4; i++)
    lastXOR ^= tmp[i];

  /*

  Transform data into hash value  by compression function using last output of transform and non-data based key

   */

  Transform(0, (u32*) & buffer[0], (u32*) & key[0].b32[0]);

  /*

  Update CRC32 of transformed data

   */

  for (i = 0; i != 15; i++)
    crc32 = ((crc32 >> 8) & 0x00FFFFFF)
            ^ crc_tab[(crc32 ^ key[0].b8[i]) & 0xFF];

}

int CalcFile(char *filename)
{

  int n;

  u8 buffer[16];
  u32 fsize;

  FILE *stream;

  if ((stream = fopen(filename, "rb+")) == NULL)
    {
      return(1);
    }

  fseek(stream, 0, SEEK_END);
  fsize = ftell(stream);
  fseek(stream, 0, SEEK_SET);

  InitHash(fsize, 0, NULL);

  do

    {

      memset(buffer, 0, 16);
      n = fread(buffer, 1, 16, stream);

      if (n != 0)
        {

          if (n != 16)
            {
              buffer[15] = n;

              if (n < 15)
                {
                  buffer[14] = 16 - n;
                }

            }

          Update(buffer);
        }

    }

  while (n == 16);

  fclose(stream);

  return(0);

}

int CalcBuffer(u8 * data, u32 len)
{

  int n, pos;

  u8 buffer[16];

  pos = 0;

  InitHash(len, 0, NULL);

  do

    {

      memset(buffer, 0, 16);

      if (len - pos >= 16)
        n = 16;
      else
        n = (len - pos);

      memmove(buffer, data + pos, n);

      pos += n;

      if (n != 0)
        {

          if (n != 16)
            {
              buffer[15] = n;

              if (n < 15)
                {
                  buffer[14] = 16 - n;
                }

            }

          Update(buffer);
        }

    }

  while (pos != len);

  return(0);

}

void PrintDigest(void)
{

  int i;

  for (i = 0; i != 16; i++)
    {

      fprintf(stdout, "%2d  0x%x \n", i, key[0].b8[i]);

    }

  fprintf(stdout, "\n\n\n");

  for (i = 0; i != 4; i++)
    {

      fprintf(stdout, " 0x%x ", (unsigned int)key[0].b32[i]);

    }

  fprintf(stdout, "\n\n\n");

}

int main(int argc, char *argv[])
{

  char text[256];

  if (argc < 2)
    {
      fprintf(stderr, "usage: %s <input filename or buffer>\n", argv[0]);
      return(1);
    }

  if (strcmp(argv[1], "buffer") == 0)
    {

      do
        {
          puts("Enter some string to calculate it's hash value:");
          memset(text, 0, 256);
          gets(text);
          fprintf(stdout, "string %s of length %d\n", text, strlen(text));
          CalcBuffer((u8*)text, strlen(text));
          PrintDigest();
        }

      while (1);

    }
  else

    {

      if (CalcFile(argv[1]) == 1)
        {
          fprintf(stderr, "failed to calculate hash value of '%s' \n",
                  argv[1]);
          return(1);
        }
      else
        {
          PrintDigest();
        }

    }
  return(0);

}
